/*
 * Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
 * Copyright (C) 2010 - DIGITEO - Allan SIMON
 * 
 * This file must be used under the terms of the CeCILL.
 * This source file is licensed as described in the file COPYING, which
 * you should have received as part of this distribution.  The terms
 * are also available at    
 * http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
 *
 */

package org.scilab.modules.metanet;

import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;
import java.util.Vector;

import javax.swing.Timer;

import org.apache.commons.logging.LogFactory;
import org.scilab.modules.graph.ScilabGraph;
import org.scilab.modules.graph.actions.CopyAction;
import org.scilab.modules.graph.actions.CutAction;
import org.scilab.modules.graph.actions.DeleteAction;
import org.scilab.modules.graph.actions.InvertSelectionAction;
import org.scilab.modules.graph.actions.PasteAction;
import org.scilab.modules.graph.actions.RedoAction;
import org.scilab.modules.graph.actions.SelectAllAction;
import org.scilab.modules.graph.actions.UndoAction;
import org.scilab.modules.graph.actions.ZoomInAction;
import org.scilab.modules.graph.actions.ZoomOutAction;
import org.scilab.modules.gui.bridge.tab.SwingScilabTab;
import org.scilab.modules.gui.menu.Menu;
import org.scilab.modules.gui.menu.ScilabMenu;
import org.scilab.modules.gui.menubar.MenuBar;
import org.scilab.modules.gui.menubar.ScilabMenuBar;
import org.scilab.modules.gui.pushbutton.PushButton;
import org.scilab.modules.gui.tab.ScilabTab;
import org.scilab.modules.gui.tab.Tab;
import org.scilab.modules.gui.textbox.ScilabTextBox;
import org.scilab.modules.gui.toolbar.ScilabToolBar;
import org.scilab.modules.gui.toolbar.ToolBar;
import org.scilab.modules.gui.utils.Size;
import org.scilab.modules.gui.window.ScilabWindow;
import org.scilab.modules.gui.window.Window;
import org.scilab.modules.metanet.actions.CircuitAction;
import org.scilab.modules.metanet.actions.CloseAction;
import org.scilab.modules.metanet.actions.CloseViewportAction;
import org.scilab.modules.metanet.actions.FitDiagramToViewAction;
import org.scilab.modules.metanet.actions.NewDiagramAction;
import org.scilab.modules.metanet.actions.OpenAction;
import org.scilab.modules.metanet.actions.QuitAction;
import org.scilab.modules.metanet.actions.RecentFileAction;
import org.scilab.modules.metanet.actions.SalesmanAction;
import org.scilab.modules.metanet.actions.SaveAction;
import org.scilab.modules.metanet.actions.SaveAsAction;
import org.scilab.modules.metanet.actions.ShortestPathAction;
import org.scilab.modules.metanet.configuration.ConfigurationManager;
import org.scilab.modules.metanet.configuration.model.DocumentType;
import org.scilab.modules.metanet.configuration.utils.ConfigurationConstants;
import org.scilab.modules.metanet.edge.BasicEdge;
import org.scilab.modules.metanet.edge.actions.AddEdgeDataFieldAction;
import org.scilab.modules.metanet.edge.actions.EditDefaultEdgeAction;
import org.scilab.modules.metanet.graph.MetanetDiagram;
import org.scilab.modules.metanet.node.BasicNode;
import org.scilab.modules.metanet.node.actions.AddNodeDataFieldAction;
import org.scilab.modules.metanet.node.actions.EditDefaultNodeAction;
import org.scilab.modules.metanet.utils.MetanetMessages;

import com.mxgraph.swing.mxGraphComponent;
import com.mxgraph.swing.mxGraphOutline;
import com.mxgraph.view.mxGraph;

public class MetanetTab  extends ScilabTab {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private static final Size WIN_SIZE = new Size(600, 500);

	/*
	 * Static fields
	 */
	private static List<MetanetDiagram> diagrams = new Vector<MetanetDiagram>();

	/*
	 * Instance fields
	 */
	private MetanetDiagram diagram;

	private MenuBar menuBar;
	private Menu fileMenu;
	private Menu recentsMenu;
	private Menu graphMenu;
	private Menu edit;
	private Menu compute;

	private PushButton openAction;
	private PushButton saveAction;
	private PushButton saveAsAction;
	private PushButton newDiagramAction;
	private PushButton deleteAction;
	private PushButton undoAction;
	private PushButton redoAction;
	private PushButton fitDiagramToViewAction;
	private PushButton zoomInAction;
	private PushButton zoomOutAction;

	/**
	 * Default constructor
	 * @param diagram The associated diagram model
	 */
	public MetanetTab(MetanetDiagram diagram) {
		super(MetanetMessages.METANET);

		this.diagram = diagram;

		initComponents();
		//this.setContentPane(diagram.getAsComponent());
		// No SimpleTab.addMember(ScilabComponent ...) so perform a raw association.
		((SwingScilabTab) getAsSimpleTab()).setContentPane(diagram.getAsComponent());

		setCallback(new CloseAction(diagram));
		diagram.getAsComponent().addKeyListener(new ArrowKeysListener());

	}

	/**
	 * Instantiate all the subcomponents of this Tab.
	 */
	private void initComponents() {
		Window window = ScilabWindow.createWindow();
		window.setDims(WIN_SIZE);

		/* Create the menu bar */
		menuBar = createMenuBar();
		addMenuBar(menuBar);

		/* Create the toolbar */
		ToolBar toolBar = createToolBar();
		addToolBar(toolBar);

		/* Create the infoBar */
		addInfoBar(ScilabTextBox.createTextBox());
		
		window.addTab(this);
	}

	/**
	 * Move cells with the arrow keys.
	 */
	private class ArrowKeysListener implements KeyListener {

		private static final double DEFAULT_PIXEL_MOVE = 1;
		private static final double MODIFIER_FACTOR = 5;
		private static final int DEFAULT_DELAY = 800; // milliseconds

		private double xIncrement;
		private double yIncrement;
		private mxGraph graph;

		private Timer repetitionTimer;
		private ActionListener doMove = new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				if (graph != null) {
					Object[] cells = graph.getSelectionCells();

					graph.getModel().beginUpdate();
					for (Object cell : cells) {
						if (cell instanceof BasicEdge
								|| cell instanceof BasicNode) {
							graph.translateCell(cell, xIncrement, yIncrement);
						}
					}
					graph.getModel().endUpdate();
					graph.refresh();
				}
			}
		};

		/**
		 * Constructor
		 */
		public ArrowKeysListener() {
			repetitionTimer = new Timer(DEFAULT_DELAY, doMove);
			repetitionTimer.setInitialDelay(0);
		}

		/**
		 * Get the action parameters and start the action timer.
		 * 
		 * @param e
		 *            key event
		 */
		public void keyPressed(KeyEvent e) {
			double realMove;
			boolean mustMove = true;

			mxGraphComponent sourceDiagram = (mxGraphComponent) e.getSource();
			graph = sourceDiagram.getGraph();

			if (graph.isGridEnabled()) {
				realMove = graph.getGridSize();
			} else {
				realMove = DEFAULT_PIXEL_MOVE;
			}

			if (e.getModifiers() == Toolkit.getDefaultToolkit()
					.getMenuShortcutKeyMask()) {
				realMove *= MODIFIER_FACTOR;
			}

			switch (e.getKeyCode()) {
			case KeyEvent.VK_UP:
				yIncrement = -realMove;
				break;

			case KeyEvent.VK_DOWN:
				yIncrement = realMove;
				break;

			case KeyEvent.VK_RIGHT:
				xIncrement = realMove;
				break;

			case KeyEvent.VK_LEFT:
				xIncrement = -realMove;
				break;

			default:
				mustMove = false;
			break;
			}

			if (!mustMove) {
				return;
			}

			if (!graph.isGridEnabled()) {
				xIncrement *= sourceDiagram.getZoomFactor();
				yIncrement *= sourceDiagram.getZoomFactor();
			}

			repetitionTimer.start();
		}

		/**
		 * Stop the action timer and clear parameters
		 * 
		 * @param e
		 *            key event
		 */
		public void keyReleased(KeyEvent e) {
			repetitionTimer.stop();
			yIncrement = 0;
			xIncrement = 0;
			graph = null;
		}

		/**
		 * Not used there
		 * 
		 * @param e
		 *            Not used
		 */
		public void keyTyped(KeyEvent e) {
		}
	}

	/**
	 * @param xcosDiagram diagram
	 */
	public static void createTabFromDiagram(MetanetDiagram metanetDiagram) {

		diagrams.add(metanetDiagram);

		MetanetTab tab = new MetanetTab(metanetDiagram);
		tab.setName(MetanetMessages.UNTITLED);
		metanetDiagram.setParentTab(tab);


		/*
		 * VIEW PORT
		 */
		MetanetTab.createViewPort(metanetDiagram);
	}

	/**
	 * Try to focus to an already openned file
	 * 
	 * @param filename filename
	 * @return True when found and focused, False otherwise
	 */
	public static boolean focusOnExistingFile(String filename) {
		for (MetanetDiagram diagram : diagrams) {
			if (diagram.getSavedFile() != null) {
				if (diagram.getSavedFile().compareTo(filename) == 0) {
					diagram.getParentTab().setCurrent();
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * @param diag diagram
	 */
	public static void closeDiagram(MetanetDiagram diag) {


		diagrams.remove(diag);

		if (diagrams.size() == 0) {
			Metanet.closeSession();
		}

	}


	/**
	 * @param MetanetDiagram diagram
	 */
	public static void showTabFromDiagram(MetanetDiagram metanetDiagram) {
		assert metanetDiagram.isOpened();

		metanetDiagram.setVisible(true);

	}

	/**
	 * Create the windows menu bar
	 * @param scilabGraph graph the associated graph
	 * @return menu bar
	 */
	private MenuBar createMenuBar() {

		menuBar = ScilabMenuBar.createMenuBar();

		/** FILE MENU */
		fileMenu = ScilabMenu.createMenu();
		fileMenu.setText(MetanetMessages.FILE);
		fileMenu.setMnemonic('F');

		fileMenu.add(NewDiagramAction.createMenu(diagram));

		
		fileMenu.add(OpenAction.createMenu(diagram));
		
		
		recentsMenu = ScilabMenu.createMenu();
		recentsMenu.setText(MetanetMessages.RECENT_FILES);
		
		final ConfigurationManager manager = ConfigurationManager.getInstance();
		final List<DocumentType> recentFiles = manager.getSettings().getRecentFiles().getDocument();
		for (int i = 0; i < recentFiles.size(); i++) {
			URL url;
			try {
				url = new URL(recentFiles.get(i).getUrl());
			} catch (MalformedURLException e) {
				LogFactory.getLog(MetanetTab.class).error(e);
				break;
			}
		    recentsMenu.add(RecentFileAction.createMenu(url));
		}
		
		ConfigurationManager.getInstance().addPropertyChangeListener(ConfigurationConstants.RECENT_FILES_CHANGED, new PropertyChangeListener() {
			@Override
			public void propertyChange(PropertyChangeEvent evt) {
				assert evt.getPropertyName().equals(ConfigurationConstants.RECENT_FILES_CHANGED);
				
				/*
				 * We only handle menu creation there. Return when this is not the case.
				 */
				if (evt.getOldValue() != null) {
					return;
				}
				
				URL url;
				try {
					url = new URL(((DocumentType)evt.getNewValue()).getUrl());
				} catch (MalformedURLException e) {
					LogFactory.getLog(MetanetTab.class).error(e);
					return;
				}
				
				recentsMenu.add(RecentFileAction.createMenu(url));
			}
		});
		
		
		fileMenu.add(recentsMenu);
		
		
		fileMenu.addSeparator();
		fileMenu.add(SaveAction.createMenu(diagram));
		fileMenu.add(SaveAsAction.createMenu(diagram));
		//fileMenu.add(ExportAction.createMenu(scilabGraph));
		fileMenu.add(CloseAction.createMenu(diagram));
		fileMenu.addSeparator();
		fileMenu.add(QuitAction.createMenu(diagram));

		menuBar.add(fileMenu);
		
		/** Graph menu*/
		
		graphMenu = ScilabMenu.createMenu();
		graphMenu.setText(MetanetMessages.GRAPH);
		graphMenu.setMnemonic('G');
		
		graphMenu.add(EditDefaultNodeAction.createMenu(diagram));
		graphMenu.add(EditDefaultEdgeAction.createMenu(diagram));
		graphMenu.add(AddEdgeDataFieldAction.createMenu(diagram));
		graphMenu.add(AddNodeDataFieldAction.createMenu(diagram));
		
		menuBar.add(graphMenu);
		
		/** Edit menu */
		edit = ScilabMenu.createMenu();
		edit.setText(MetanetMessages.EDIT);
		edit.setMnemonic('E');
		menuBar.add(edit);

		edit.add(UndoAction.undoMenu(diagram));
		edit.add(RedoAction.redoMenu(diagram));
		edit.addSeparator();
		edit.add(CutAction.cutMenu(diagram));
		edit.add(CopyAction.copyMenu(diagram));
		edit.add(PasteAction.pasteMenu(diagram));
		edit.add(DeleteAction.createMenu(diagram));
		edit.addSeparator();
		edit.add(SelectAllAction.createMenu(diagram));
		edit.add(InvertSelectionAction.createMenu(diagram));
		edit.addSeparator();
		//edit.add(BlockParametersAction.createMenu(diagram));
		

		menuBar.add(edit);
		
		/** Compute Menu */
		compute = ScilabMenu.createMenu();
		compute .setText(MetanetMessages.COMPUTE);
		compute.setMnemonic('C');
		menuBar.add(compute);

		compute.add(ShortestPathAction.createMenu(diagram));
		compute.add(CircuitAction.createMenu(diagram));
		compute.add(SalesmanAction.createMenu(diagram));
		
		return menuBar;

	}

	/**
	 * @return tool bar
	 */
	private ToolBar createToolBar() {
		ToolBar toolBar = ScilabToolBar.createToolBar();

		
		newDiagramAction = NewDiagramAction.createButton(diagram);
		toolBar.add(newDiagramAction);
		

		openAction = OpenAction.createButton(diagram);
		toolBar.add(openAction);

		toolBar.addSeparator();

		saveAction = SaveAction.createButton(diagram);
		toolBar.add(saveAction);
		saveAsAction = SaveAsAction.createButton(diagram);
		toolBar.add(saveAsAction);

		toolBar.addSeparator();

		/*
		printAction = PrintAction.createButton(diagram);
		toolBar.add(printAction);
		*/

		toolBar.addSeparator();

		deleteAction = DeleteAction.createButton(diagram);
		toolBar.add(deleteAction);

		toolBar.addSeparator();

		// UNDO / REDO
		undoAction = UndoAction.undoButton(diagram);
		redoAction = RedoAction.redoButton(diagram);
		toolBar.add(undoAction);
		toolBar.add(redoAction);

		toolBar.addSeparator();

		fitDiagramToViewAction = FitDiagramToViewAction.createButton(diagram);
		toolBar.add(fitDiagramToViewAction);

		toolBar.addSeparator();

		// ZOOMIN / ZOOMOUT
		zoomInAction = ZoomInAction.zoominButton(diagram);
		toolBar.add(zoomInAction);
		zoomOutAction = ZoomOutAction.zoomoutButton(diagram);
		toolBar.add(zoomOutAction);

		return toolBar;
	}

	/**
	 * @param xcosDiagramm diagram
	 */
	public static void createViewPort(ScilabGraph metanetDiagramm) {
		Window outline = ScilabWindow.createWindow();
		Tab outlineTab = ScilabTab.createTab(MetanetMessages.VIEWPORT);

		outlineTab.setCallback(new CloseViewportAction(metanetDiagramm));

		MenuBar vpMenuBar = ScilabMenuBar.createMenuBar();
		outlineTab.addMenuBar(vpMenuBar);

		Menu vpMenu = ScilabMenu.createMenu();
		vpMenu.setText(MetanetMessages.VIEWPORT);
		vpMenu.setMnemonic('V');
		vpMenuBar.add(vpMenu);

		vpMenu.add(CloseViewportAction.createMenu(metanetDiagramm));

		outlineTab.getAsSimpleTab().setInfoBar(ScilabTextBox.createTextBox());

		((MetanetDiagram) metanetDiagramm).setViewPort(outlineTab);

		// Creates the graph outline component
		mxGraphOutline graphOutline = new mxGraphOutline(metanetDiagramm
				.getAsComponent());

		graphOutline.setDrawLabels(true);

		((SwingScilabTab) outlineTab.getAsSimpleTab())
		.setContentPane(graphOutline);
		outline.addTab(outlineTab);
		outline.setVisible(false);
		outlineTab.setVisible(false);
	}

	/**
	 * @return diagram list
	 */
	public static List<MetanetDiagram> getAllDiagrams() {
		return diagrams;
	}

}

